# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import pathlib
from collections import defaultdict
from typing import Optional

import libcst as cst
import pandas as pd
from libcst.metadata import ParentNodeProvider, PositionProvider
from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.training.dataflowmodel import ast_operation as ast

logger = setup_logger()

sig2label = {}


class DependentApiCollector(cst.CSTVisitor):
    """DependentApiCollector class

    This script groups the APIs that are dependent on each other based on the given pipelins in the corpus.
    In SapientML, two APIs: A and B are dependent on each other if both A and B are applied on at least one column
    of the dataset in the given pipeline.

    """

    METADATA_DEPENDENCIES = (
        ParentNodeProvider,
        PositionProvider,
    )

    def __init__(self, relative_file_path, columns):
        self.as_names = {}
        self.columns = columns
        self.column2apis = {}
        self.user_defined_function = False
        self.relative_file_path = relative_file_path
        self.count = 0

    def visit_SimpleString(self, node: cst.SimpleString) -> Optional[bool]:
        """visit_SimpleString function

        Parameters
        ----------
        node : SimpleString
            It is libcst node class

        """
        string_value = node.value

        string_without_quote = string_value.replace("'", "")
        string_without_quote = string_without_quote.replace('"', "")

        if string_without_quote in self.columns:
            parent = self.get_metadata(ParentNodeProvider, node)
            line_number = -1
            while not (isinstance(parent, cst.SimpleStatementLine) or isinstance(parent, cst.Module)):
                if isinstance(parent, cst.Call):
                    pos = self.get_metadata(PositionProvider, parent).start
                    line_number = pos.line
                    break
                parent = self.get_metadata(ParentNodeProvider, parent)

            if line_number != -1:
                sig = self.relative_file_path + ":" + str(line_number)
                current_list = []
                if string_without_quote in self.column2apis.keys():
                    current_list = self.column2apis[string_without_quote]

                label = sig2label[sig] if sig in sig2label.keys() else "N/A"

                if label not in current_list and label != "N/A":
                    current_list.append(label)

                self.column2apis[string_without_quote] = current_list

    def visit_Assign(self, node: cst.Assign) -> Optional[bool]:
        """visit_Assign function

        Parameters
        ----------
        node : Assign
            It is libcst node class

        """
        try:
            assigned_target = node.targets[0]
            target = assigned_target.target
            check = hasattr(target, "value")
            if check and target.value not in self.as_names:
                value = node.value
                if isinstance(value, cst.Call):
                    func = value.func
                    if isinstance(func, cst.Name):
                        name = func.value
                    elif isinstance(func, cst.Attribute):
                        name = func.attr.value
                    self.as_names[target.value] = name
                else:
                    self.count += 1
                    self.as_names[target.value] = "@@VAR@@"
        except Exception:
            pass

    def visit_ImportAlias(self, node: cst.ImportAlias):
        """visit_ImportAlias function

        Parameters
        ----------
        node : ImportAlias
            It is libcst node class

        """
        if node.asname is not None:
            type_str = ""
            child = node.name
            while not isinstance(child, cst.Name):
                type_str = type_str + "_" + child.attr.value
                child = child.value
            type_str = type_str + "_" + child.value
            self.as_names[str(node.asname.name.value)] = str(type_str)

    def get_Call_info(self, node):
        """get_Call_info function

        Returns
        -------
        function_call : str

        """
        if isinstance(node.func, cst.Attribute):  # c.m(...)
            function_call = self.getCallFullyQualifiedName(node.func)
        elif isinstance(node.func, cst.Name):  # m(...)
            function_name = node.func.value
            function_call = function_name
        return function_call

    def getCallFullyQualifiedName(self, func_node):
        """getCallFullyQualifiedName function

        Returns
        -------
        str
            It returns fully qualified name

        """
        if isinstance(func_node, cst.Name):  # recursive call may trigger this condition
            return func_node.value
        if isinstance(func_node.value, cst.Name):
            module_name = func_node.value.value
            function_name = func_node.attr.value
            function_call = (
                (self.as_names[module_name] if module_name in self.as_names.keys() else module_name)
                + "."
                + function_name
            )
            return function_call
        elif isinstance(func_node.value, cst.Subscript):
            module_name = "UNKNOWN_TYPE"
            function_name = func_node.attr.value
            function_call = module_name + "." + function_name
            return function_call
        elif isinstance(func_node.value, cst.Call):
            function_name = func_node.attr.value
            return self.getCallFullyQualifiedName(func_node.value.func) + "." + function_name
        elif isinstance(func_node.value, cst.Attribute):
            function_name = func_node.attr.value
            if isinstance(func_node.value.value, cst.Name):
                return func_node.value.value.value + "." + function_name
            elif isinstance(func_node.value.value, cst.Call):
                attr_name = func_node.value.attr.value
                return (
                    self.getCallFullyQualifiedName(func_node.value.value.func) + "." + attr_name + "." + function_name
                )
            elif isinstance(func_node.value.value, cst.Subscript):
                attr_name = func_node.value.attr.value
                return "UNKNOWN_TYPE" + "." + attr_name + "." + function_name
            elif isinstance(func_node.value.value, cst.Attribute):
                return self.getAttributeFullyQualifiedName(func_node.value) + "." + function_name
            else:
                return "XXX"
        elif isinstance(func_node.value, cst.SimpleString):  # e.g., "".join()
            return "str." + func_node.attr.value
        elif isinstance(func_node.value, cst.Comparison):  # e.g., (z > 3).xxx()
            return "bool." + func_node.attr.value
        elif isinstance(func_node.value, cst.BinaryOperation):  # e.g., (a - b).min()
            return "bin_op." + func_node.attr.value
        else:
            return "XXX"

    def getAttributeFullyQualifiedName(self, func_node):
        """getAttributeFullyQualifiedName function

        Returns
        -------
        str
            It returns fully qualified attribute name.

        """
        function_name = func_node.attr.value
        if isinstance(func_node.value.value, cst.Name):
            attr_name = func_node.value.attr.value
            return func_node.value.value.value + "." + attr_name + "." + function_name
        elif isinstance(func_node.value.value, cst.Call):
            attr_name = func_node.value.attr.value
            return self.getCallFullyQualifiedName(func_node.value.value.func) + "." + attr_name + "." + function_name
        elif isinstance(func_node.value.value, cst.Subscript):
            attr_name = func_node.value.attr.value
            return "UNKNOWN_TYPE" + "." + attr_name + "." + function_name
        elif isinstance(func_node.value.value, cst.Attribute):
            attr_name = func_node.value.attr.value
            return self.getAttributeFullyQualifiedName(func_node.value) + "." + attr_name + "." + function_name
        else:
            return "XXX"


def find_column_to_apis(file_path, dataset_path):
    """Extracting API's used on columns.

    Parameters
    ----------
    file_path : str
        It contains file path in string format.
    dataset_path : PosixPath
        It contains path from pathlib.PosixPath class.

    Returns
    -------
    dict
        It returns collector.column2apis in dictionary format.

    """
    source_tree = ast.construct_tree(file_path)
    data = pd.read_csv(dataset_path, encoding="ISO-8859-1")
    parts = file_path.split("/")
    relative_file_path = parts[-1]
    collector = DependentApiCollector(relative_file_path, data.columns)
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(collector)
    return collector.column2apis


def get_dataset_folder_name(file_path):
    notebook_info_path = file_path.with_suffix(".info.json")
    dataset_folder_name = ""
    try:
        with open(notebook_info_path, "r", encoding="utf-8") as notebook_info_file:
            notebook_info = json.load(notebook_info_file)
            if isinstance(notebook_info, list):
                notebook_info = notebook_info[1]

            if isinstance(notebook_info, dict):
                dataset_folder_name = notebook_info["dataset_folder"]
    except Exception:
        logger.warning("Could not read JSON info file: {}".format(notebook_info_path))
    return dataset_folder_name


def get_dataset_path(file_path):
    """Fetches the dataset path for the given pipeline.

    Parameters
    ----------
    file_path : PosixPath
        It contains path from pathlib.PosixPath

    Returns
    -------
    target_csv_file : PosixPath
        It contains path of the target_csv_file from pathlib.PosixPath class

    """
    dataset_folder_name = get_dataset_folder_name(file_path)
    parent = file_path.parent / dataset_folder_name
    csv_file_lists = pathlib.Path(str(parent).replace(internal_path.clean_notebooks_dir_name, "dataset")).rglob("*.csv")
    target_csv_file = None
    for csv_file in csv_file_lists:
        target_csv_file = csv_file
    return target_csv_file


def main():
    """Extract API's from pipelines and create dependent API's map."""
    annotated_notebooks = internal_path.project_labels_path

    label_data = pd.read_csv(annotated_notebooks, encoding="utf-8", usecols=["file_name", "new_label", "line_number"])

    for _, row in label_data.iterrows():
        key = str(row["file_name"]) + ":" + str(row["line_number"])
        new_label = str(row["new_label"])
        if new_label.startswith("PREPROCESS"):
            sig2label[key] = new_label.split("#")[0]

    final_dependency_list = defaultdict(int)

    path_list = pathlib.Path(internal_path.clean_dir).rglob("*.py")
    for path in path_list:
        logger.info(f"FILE:{path}")
        try:
            file_path = str(path)
            dataset_path = get_dataset_path(path)
            col2apis = find_column_to_apis(file_path, dataset_path)
            logger.info(col2apis)
            for key in col2apis:
                if len(col2apis[key]) > 1:
                    new_key = str(col2apis[key])
                    final_dependency_list[new_key] += 1
        except Exception:
            import traceback

            traceback.print_exc()

    sorted_map = {k: v for k, v in sorted(final_dependency_list.items(), key=lambda item: item[1], reverse=True)}

    with open(internal_path.training_cache / "dependent_labels.json", "w", encoding="utf-8") as outfile:
        json.dump(sorted_map, outfile, indent=4)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag

    main()



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import defaultdict
from typing import Literal, Optional

import numpy as np
import pandas as pd
from sapientml.util.logging import setup_logger

warnings.filterwarnings("ignore")
logger = setup_logger()


class AutoVisualization_Class:
    """AutoVisualization_Class."""

    def __init__(self):
        self.code_json = defaultdict(list)

    def AutoVisualization(
        self,
        df: pd.DataFrame,
        target_columns: list[str],
        problem_type: Literal["regression", "classification"],
        ignore_columns: Optional[list[str]] = None,
    ):
        """
        The main processing class for visualizations.

        Parameters
        ----------
        target_columns : list[str]
            Names of target columns.
        problem_type : Literal["regression", "classification"]
            Type of problem either regression or classification
        ignore_columns : list[str], optional
            Column names which must not be used and must be dropped.

        Returns
        -------
        generates visualization code.

        """
        problem_type = problem_type.lower()

        if not ignore_columns:
            ignore_columns = []

        if problem_type != "regression" and problem_type != "classification":
            logger.info(
                "Problem type should be configured either regression or classification. Please check your task specification"
            )
            # if problem type is not specified as either 'regression' or 'classification', set it to default value 'regression'
            problem_type = "regression"

        if not isinstance(target_columns, list):
            logger.error("Target columns should be defined as a list. Please check your task specification")
            return self.code_json

        if len(target_columns) == 0:
            logger.error("Target column is not configured. Please check your task specification")
            return self.code_json

        if not isinstance(ignore_columns, list):
            logger.info("ID columns should be defined as a list. Please check your task specification")

        numvars_list, catvars_list, textvars_list = self.classify_columns_coarse_granularity(
            df, ignore_columns, target_columns
        )

        numvars_list = sorted(numvars_list)
        catvars_list = sorted(catvars_list)
        textvars_list = sorted(textvars_list)
        target_columns = sorted(target_columns)

        try:
            self.draw_distplot(problem_type, numvars_list, catvars_list, target_columns)
        except Exception:
            logger.error("Could not draw Distribution Plots")

        try:
            self.draw_heatmap(problem_type, numvars_list, catvars_list, target_columns)
        except Exception:
            logger.error("Could not draw Heat Maps")

        logger.info("VISUALIZATION Completed.")

        return self.code_json

    def draw_heatmap(self, problem_type, numvars_list=None, catvars_list=None, targetvars_list=None):
        """draw_heatmap method.

        generate codes for feature heatmap.

        Parameters
        ----------
        problem_type : Literal["regression", "classification"]
            Type of problem either regression or classification
        numvars_list : List, default None
            list of numerical columns
        catvars_list : List, default None
            list of categorical columns
        targetvars_list : List, default None
            list of target columns.

        """
        total_numvars = len(numvars_list)
        total_catvars = len(catvars_list)

        max_vis_to_show = 20  # set a upper limits

        heatmap_list = (numvars_list + catvars_list)[:max_vis_to_show]
        heatmap_list.extend(targetvars_list)

        if total_numvars + total_catvars > 1:
            # generate the codes
            codes = [
                "_COLS_FOR_HEATMAP = %s" % str(heatmap_list),
                "plt.figure(figsize=(20,20))",
                'plt.title("Pearson Correlation HeatMap of Features")',
                "sns.heatmap(train_dataset[_COLS_FOR_HEATMAP].corr(numeric_only=True),annot=True)",
            ]

            self.code_json["heatmap"] = codes

    def draw_distplot(self, problem_type, numvars_list=None, catvars_list=None, targetvars_list=None):
        """
        draw_distplot method.

        generate codes for data distribution plots.

        Parameters
        ----------
        problem_type : Literal["regression", "classification"]
            Type of problem either regression or classification
        numvars_list : List | None, default None
            list of numerical columns.
        catvars_list : List | None, default None
            list of categorical columns.
        targetvars_list : List | None, default None
            list of target columns.

        """
        total_numvars = len(numvars_list)
        total_catvars = len(catvars_list)
        total_targetvars = len(targetvars_list)

        max_vis_to_show = 100  # set a upper limits

        if total_numvars >= max_vis_to_show:
            numvars_list = numvars_list[:max_vis_to_show]
            total_numvars = 100
        if total_catvars >= max_vis_to_show:
            catvars_list = catvars_list[:max_vis_to_show]
            total_catvars = 100
        if total_targetvars >= max_vis_to_show:
            targetvars_list = targetvars_list[:max_vis_to_show]
            total_targetvars = 100

        # best layout for numerical columns
        numvars_n_row, numvars_n_col, numvars_n_unvisible = 0, 0, 0
        if total_numvars <= 4:
            numvars_n_row, numvars_n_col = 1, total_numvars
        else:
            numvars_n_row, numvars_n_col = total_numvars // 4, 4
            if total_numvars % 4 != 0:
                numvars_n_row += 1  # need one more row
                numvars_n_unvisible = 4 - total_numvars % 4

        # best layout for categorical columns
        catvars_n_row, catvars_n_col, catvars_n_unvisible = 0, 0, 0
        if total_catvars <= 4:
            catvars_n_row, catvars_n_col = 1, total_catvars
        else:
            catvars_n_row, catvars_n_col = total_catvars // 4, 4
            if total_catvars % 4 != 0:
                catvars_n_row += 1  # need one more row
                catvars_n_unvisible = 4 - total_catvars % 4

        # best layout for target columns
        targetvars_n_row, targetvars_n_col, targetvars_n_unvisible = 0, 0, 0
        if total_targetvars <= 4:
            targetvars_n_row, targetvars_n_col = 1, total_targetvars
        else:
            targetvars_n_row, targetvars_n_col = total_targetvars // 4, 4
            if total_targetvars % 4 != 0:
                targetvars_n_row += 1  # need one more row
                targetvars_n_unvisible = 4 - total_targetvars % 4

        # generate the codes
        codes = [
            "import matplotlib.pyplot as plt",
            "import japanize_matplotlib",
            "import seaborn as sns",
            "import numpy as np",
            "\n",
        ]

        # add numerical columns related codes
        if total_numvars > 1:
            codes.extend(
                [
                    "fig, axes = plt.subplots(%d, %d, figsize=(%d, %d))"
                    % (numvars_n_row, numvars_n_col, 16, numvars_n_row * 4),
                    "axes = axes.ravel()",
                    "for index,col in enumerate(" + str(numvars_list) + "):",
                    "\tsns.histplot(train_dataset[col],ax=axes[index])",
                    '\taxes[index].tick_params(axis="x", rotation=90)',
                    '\taxes[index].set_title("Distribution of %s" %col )',
                    "\tif len(axes[index].get_xticks()) > 20:",
                    "\t\txtick_hidden = len(axes[index].get_xticks())//20",
                    "\t\taxes[index].set_xticks(axes[index].get_xticks()[::xtick_hidden])",
                ]
            )

            if numvars_n_unvisible > 1:
                codes.extend(
                    [
                        "for index in range(%d,0):" % (-numvars_n_unvisible),
                        "\taxes[index].set_visible(False)",
                        "fig.tight_layout()",
                        "\n",
                    ]
                )
            elif numvars_n_unvisible == 1:
                codes.extend(["axes[-1].set_visible(False)", "fig.tight_layout()", "\n"])
            else:
                codes.extend(["fig.tight_layout()", "\n"])

        if total_numvars == 1:
            codes.extend(
                [
                    "fig = plt.figure(figsize=(4,4))",
                    'sns.histplot(x = train_dataset["' + str(numvars_list[0]) + '"])',
                    "plt.xticks(rotation=90)",
                    "plt.locator_params(nbins=20)",
                    'plt.title("Distribution of %s")' % str(numvars_list[0]),
                    "fig.tight_layout()" "\n",
                ]
            )

        # add categorical columns related codes
        if total_catvars > 1:
            codes.extend(
                [
                    "fig, axes = plt.subplots(%d, %d, figsize=(%d, %d))"
                    % (catvars_n_row, catvars_n_col, 16, catvars_n_row * 4),
                    "axes = axes.ravel()",
                    "for index,col in enumerate(" + str(catvars_list) + "):",
                    '\tsns.countplot(x = train_dataset[col].fillna("").astype(str), ax=axes[index], order=sorted(train_dataset[col].fillna("").astype(str).unique()))',
                    '\taxes[index].tick_params(axis="x", rotation=90)',
                    '\taxes[index].set_title("Distribution of %s" %col )',
                    "\tif len(axes[index].get_xticks()) > 20:",
                    "\t\txtick_hidden = len(axes[index].get_xticks())//20",
                    "\t\taxes[index].set_xticks(axes[index].get_xticks()[::xtick_hidden])",
                ]
            )

            if catvars_n_unvisible > 1:
                codes.extend(
                    [
                        "for index in range(%d,0):" % (-catvars_n_unvisible),
                        "\taxes[index].set_visible(False)",
                        "fig.tight_layout()",
                        "\n",
                    ]
                )
            elif catvars_n_unvisible == 1:
                codes.extend(["axes[-1].set_visible(False)", "fig.tight_layout()", "\n"])
            else:
                codes.extend(["fig.tight_layout()", "\n"])

        if total_catvars == 1:
            codes.extend(
                [
                    "fig = plt.figure(figsize=(4,4))",
                    'sns.countplot(x = train_dataset["'
                    + str(catvars_list[0])
                    + '"].fillna("").astype(str), order=sorted(train_dataset["'
                    + str(catvars_list[0])
                    + '"].fillna("").astype(str).unique()))',
                    "plt.xticks(rotation=90)",
                    "plt.locator_params(nbins=20)",
                    'plt.title("Distribution of %s")' % str(catvars_list[0]),
                    "fig.tight_layout()" "\n",
                ]
            )

        # add target columns related codes
        if total_targetvars > 1:
            codes.extend(
                [
                    "fig, axes = plt.subplots(%d, %d, figsize=(%d, %d))"
                    % (targetvars_n_row, targetvars_n_col, 16, targetvars_n_row * 4),
                    "axes = axes.ravel()",
                    "for index,col in enumerate(" + str(targetvars_list) + "):",
                ]
            )

            if problem_type == "regression":
                codes.extend(
                    [
                        "\tsns.histplot(x = train_dataset[col],ax=axes[index])",
                        '\taxes[index].tick_params(axis="x", rotation=90)',
                        '\taxes[index].set_title("Distribution for Target Column of %s" %col )',
                    ]
                )
            else:
                # NOTE: In version 2.6.9, multi-target classification doesn't working.
                #       So, this route can't be tested.
                codes.extend(
                    [
                        '\tsns.countplot(x = train_dataset[col].fillna("").astype(str), ax=axes[index], order=sorted(train_dataset[col].fillna("").astype(str).unique()))',
                        '\taxes[index].tick_params(axis="x", rotation=90)',
                        '\taxes[index].set_title("Distribution for Target Column of %s" %col )',
                    ]
                )

            if targetvars_n_unvisible > 1:
                codes.extend(
                    [
                        "for index in range(%d,0):" % (-targetvars_n_unvisible),
                        "\taxes[index].set_visible(False)",
                        "fig.tight_layout()",
                        "\n",
                    ]
                )
            elif targetvars_n_unvisible == 1:
                codes.extend(["axes[-1].set_visible(False)", "fig.tight_layout()", "\n"])
            else:
                codes.extend(["fig.tight_layout()", "\n"])

        if total_targetvars == 1:
            codes.extend(
                [
                    "fig = plt.figure(figsize=(5,5))",
                ]
            )

            if problem_type == "regression":
                codes.extend(
                    [
                        'sns.histplot(x = train_dataset["' + str(targetvars_list[0]) + '"])',
                    ]
                )
            else:
                codes.extend(
                    [
                        'sns.countplot(x = train_dataset["'
                        + str(targetvars_list[0])
                        + '"].fillna("").astype(str), order=sorted(train_dataset["'
                        + str(targetvars_list[0])
                        + '"].fillna("").astype(str).unique()))',
                    ]
                )

            codes.extend(
                [
                    "plt.xticks(rotation=90)",
                    'plt.title("Distribution for Target Column of %s")' % str(targetvars_list[0]),
                    "fig.tight_layout()" "\n",
                ]
            )

        self.code_json["distplot"] = codes

    def classify_columns_coarse_granularity(self, df_preds, ignore_columns, target_columns, verbose=3):
        """
        classify_columns_coarse_granularity method.

        simple classification function for columns.
        only to classify the columns into numerical, category and text columns.

        Parameters
        ----------
        df_preds : pd.DataFrame
            input dataframe
        ignore_columns : list[str], optional
            Column names which must not be used and must be dropped.
        target_columns : list[str]
            Names of target columns.

        Results
        -------
        numvars_list : List
            list of numerical columns
        catvars_list : List
            list of categorical columns
        textvars_list : List
            list of text columns.

        """
        # categorical column threshold
        cat_threshold = 0.05
        catvars_list = []
        numvars_list = []
        textvars_list = []

        for var in df_preds.columns:
            # add .astype(str) to avoid error if dataframe has list values when caluculate nunique.
            if 1.0 * df_preds[var].fillna("").astype(str).nunique() / df_preds[var].count() < cat_threshold:
                if (
                    var not in ignore_columns and var not in target_columns
                ):  # id column and target column need to be removed from consideration
                    catvars_list.append(var)

        object_vars_all = df_preds.select_dtypes(include="object").columns.tolist()
        num_vars_all = df_preds.select_dtypes(include=np.number).columns.tolist()

        for var in num_vars_all:
            if var not in catvars_list:
                if (
                    var not in ignore_columns and var not in target_columns
                ):  # id column and target column need to be removed from consideration
                    numvars_list.append(var)

        for var in object_vars_all:
            if var not in catvars_list:
                if (
                    var not in ignore_columns and var not in target_columns
                ):  # id column and target column need to be removed from consideration
                    textvars_list.append(var)

        return numvars_list, catvars_list, textvars_list



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pickle
from collections import OrderedDict, defaultdict
from typing import Literal

import pandas as pd
from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.design import search_space
from sapientml_core.training import meta_feature_selector
from sklearn.tree import DecisionTreeClassifier

logger = setup_logger()


def train_p_model(X, y):
    """Build a decision tree classifier from the training set (X, y).

    Parameters
    ----------
    X : MatrixLike = np.ndarray | pd.DataFrame | spmatrix   |
        ArrayLike = numpy.typing.ArrayLike
        The training input samples.
    y : MatrixLike = np.ndarray | pd.DataFrame | spmatrix   |
        ArrayLike = numpy.typing.ArrayLike
        The target values (class labels) as integers or strings

    Returns
    -------
    model : DecisionTreeClassifier
        Fitted estimator.
    """
    model = DecisionTreeClassifier(class_weight="balanced", max_depth=3)
    model.fit(X, y)
    return model


def _train_preprocessors(train_data, feature_selection: Literal["select_manually", "customized"]):
    logger.info("Training skeleton predictor for preprocessors...")
    data = train_data
    data.drop(
        data.filter(regex="(TEMPLATE|IGNORE|EVAL:|RPEPROCESS:|MODEL:|Unnamed:)").columns,
        axis=1,
        inplace=True,
    )
    data["project_target"] = (
        data["csv_name"] + "_" + data["target_column_name"].apply(lambda line: "_".join(sorted(eval(line))))
    )
    all_labels = [v for v in data.columns if v.startswith(("PREPROCESS:"))]
    second_to_full_labels = defaultdict(list)
    for label in all_labels:
        second_to_full_labels["PREPROCESS:" + label.split(":")[1]].append(label)

    pp_models = OrderedDict()

    selected_features_map = meta_feature_selector.select_based_on_correlation(data)

    for _, detail_labels in second_to_full_labels.items():
        for label in detail_labels:
            logger.debug(label)
            main_df = data.copy()
            # Feature Selection On
            y = main_df[label]
            X = main_df[search_space.meta_feature_list]

            if feature_selection == "select_manually":
                selected_features = meta_feature_selector.select_features(label)
                logger.debug("Selected Features:", selected_features)
                X = main_df[selected_features]
            elif feature_selection == "customized":
                selected_features = selected_features_map[label]
                if len(selected_features) == 0:
                    selected_features = meta_feature_selector.select_sequentially(X, y)
                logger.debug("Selected Features:", selected_features)
                X = main_df[selected_features]

            pp_model = train_p_model(X, y)
            pp_models[label] = (pp_model, selected_features)

    return pp_models


def _prepare_model_training_data(raw_meta_feature_train):
    # Remove all the unnecessary meta-features
    final_meta_features = raw_meta_feature_train[search_space.project_related_metadata + search_space.meta_feature_list]
    final_meta_features.fillna(0, inplace=True)
    for semantic_label, columns in search_space.label_mapping.items():
        try:
            final_meta_features[semantic_label] = raw_meta_feature_train[columns].sum(axis=1)
            final_meta_features[semantic_label] = final_meta_features[semantic_label].apply(lambda x: 1 if x > 0 else 0)
        except KeyError as e:
            logger.warning(e)

    return final_meta_features


def main():
    """This main function preprocesses the learning data and saves fitted estimator for the DecisionTreeClassifier.

    Description of feature_selection : "select_manually" | "customized"
        Specify how features are selected.
    """
    training_data_path = internal_path.training_cache / "pp_metafeatures_training.csv"
    # "select_manually" | "customized"
    feature_selection = "customized"
    raw_meta_feature_train = pd.read_csv(training_data_path)
    meta_feature_train = _prepare_model_training_data(raw_meta_feature_train)
    pp_models = _train_preprocessors(meta_feature_train, feature_selection)
    # Save model
    with open(internal_path.training_cache / "pp_models.pkl", "wb") as f:
        pickle.dump(pp_models, f)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag

    main()



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
from collections import defaultdict
from pathlib import Path
from typing import Optional

from sapientml import macros
from sapientml.params import Config, Task
from sapientml.util.logging import setup_logger

from ...params import DatasetSummary, ModelLabel, Pipeline, PipelineSkeleton, SapientMLConfig, SimplePipeline
from .pipeline_template import PipelineTemplate, is_allowed_to_apply_to_target
from .preprocessing_label import PreprocessingLabel

logger = setup_logger()

preprocessing_threshold = 0.5


class Adaptation:
    """A class to represent the adaptation.

    This class provides core functionalities for the adaptation module.
    More specifically, it
    i) identify relevant columns in the given dataset for each ML component where the component will be instantiated on.
    ii) removes redundant ML components
    iii) order them properly
    iv) finally assemble all the components together.

    Parameters
    ----------
    pipeline : pd.DataFrame

    """

    pipeline: PipelineTemplate

    PREPROCESSING_AFTER_TARGET_SEPARATION = {
        "PREPROCESS:Category:get_dummies:pandas",
        "PREPROCESS:Scaling:StandardScaler:sklearn",
        "PREPROCESS:Text:CountVectorizer:sklearn",
        "PREPROCESS:Text:TfidfVectorizer:sklearn",
        "PREPROCESS:Scaling:STANDARD:sklearn",
        "PREPROCESS:Scaling:StandardScaler:sklearn",
    }

    PREPROCESSING_AFTER_TRAIN_TEST_SPLIT = {"PREPROCESS:Balancing:SMOTE:imblearn"}

    PREPROCESSING_ONLY_TARGET = {"PREPROCESS:Scaling:log:custom", "PREPROCESS:Balancing:SMOTE:imblearn"}

    def __init__(
        self,
        labels: PipelineSkeleton,
        task: Task,
        dataset_summary: DatasetSummary,
        config: Config,
    ):
        """Constructs all the necessary attributes for the adaptation object.

        Parameters
        ----------
        labels : dict
            Preprocess and model components details.
        task : Task
            Object of the task class.
        dataset_summary : DatasetSummary
            Object of the datasetsummary class.
        config : Config
            Object of the config class.

        """
        self.labels = labels
        self.task = task
        self.dataset_summary = dataset_summary
        self.config = config
        self.adaptation_metric: Optional[str] = self._get_adaptation_metric_label()

        # Load all the training Data
        with open(Path(os.path.dirname(__file__)) / "../artifacts/label_order.json", "r", encoding="utf-8") as f:
            label_order = json.load(f)

        self.label_order = label_order
        self.dataset_summary = dataset_summary
        self.pipeline = PipelineTemplate(
            pipeline=Pipeline(
                task=self.task,
                dataset_summary=self.dataset_summary,
                config=SapientMLConfig(**self.config.model_dump()),
            )
        )

    def _get_labels_from_skeleton_predictor(self):
        # fetch preprocessing and model labels
        all_preprocessing_labels = dict()
        model_labels = dict()
        for label_name, details in self.labels.items():
            if label_name.startswith("PREPROCESS"):
                all_preprocessing_labels[label_name] = details
            else:
                model_labels[label_name] = details

        # filter labels based on config setting
        preprocessing_labels = dict(
            filter(
                lambda elem: elem[1]["probability"] > preprocessing_threshold,
                all_preprocessing_labels.items(),
            )
        )

        # resolve conflicting preprocessing labels by keeping the one with highest probability
        conflicting_labels = dict()
        preprocessing_labels, conflicting_labels = self._resolve_conflicting_labels(preprocessing_labels)

        # remove GausianNB and SVC if TfidfVectorizer is contain
        if "PREPROCESS:Text:TfidfVectorizer:sklearn" in preprocessing_labels:
            model_labels = {k: v for k, v in model_labels.items() if "GaussianNB" not in k and "SVC" not in k}

        n_models = self.config.n_models
        if n_models < 1:
            raise ValueError("Please set 'n_models' to a number greater than or equal to 1.")
        model_labels = dict(list(model_labels.items())[0:n_models])

        # order preprocessing labels and model labels
        preprocessing_labels, model_labels = self._order_labels(preprocessing_labels, model_labels)

        # convert to preprocessing and model objects
        preprocessing_label_objects: list[PreprocessingLabel] = list()
        model_label_objects: list[ModelLabel] = list()
        for name, details in preprocessing_labels.items():
            pObj = PreprocessingLabel(name, details["features"], details["predicates"])
            labels_before, labels_after = self._get_preprocessing_labels_before_and_after(name, preprocessing_labels)
            pObj.components_before = labels_before
            pObj.components_after = labels_after
            if name in conflicting_labels:
                for label in conflicting_labels[name]:
                    pObj.alternative_components.append(
                        PreprocessingLabel(
                            label,
                            all_preprocessing_labels[label]["features"],
                            all_preprocessing_labels[label]["predicates"],
                        )
                    )

            preprocessing_label_objects.append(pObj)

        for name, details in model_labels.items():
            mObj = ModelLabel(label_name=name)
            model_label_objects.append(mObj)

        return preprocessing_label_objects, model_label_objects

    def _get_preprocessing_labels_before_and_after(self, label, preprocessing_labels):
        labels_before = list()
        labels_after = list()
        for combination in self.label_order:
            arr = combination.split("#")
            first = arr[0]
            second = arr[1]
            if label == second and first in preprocessing_labels:
                labels_before.append(first)
            elif label == first and second in preprocessing_labels:
                labels_after.append(second)
        return labels_before, labels_after

    def _resolve_conflicting_labels(self, preprocessing_labels):
        second_level_labels_dict = defaultdict(list)
        for label in preprocessing_labels:
            second_level = label.split(":")[1]
            second_level_labels_dict[second_level].append(label)
        clean_labels = dict()
        conflicting_labels = dict()
        for second_level in second_level_labels_dict:
            sorted_list = sorted(
                second_level_labels_dict[second_level],
                key=lambda i: preprocessing_labels[i]["probability"],
                reverse=True,
            )
            label_to_keep = sorted_list[0]
            clean_labels[label_to_keep] = preprocessing_labels[label_to_keep]

            if len(sorted_list) > 1:
                conflicting_labels[sorted_list[0]] = sorted_list[1:]

        preprocessing_labels = clean_labels
        return preprocessing_labels, conflicting_labels

    def _order_labels(self, preprocessing_labels, model_labels):
        # sort preprocessing components
        sorted_label_names = self._sort(list(preprocessing_labels.keys()), self.label_order)
        sorted_preprocessing_labels = dict()
        for label in sorted_label_names:
            sorted_preprocessing_labels[label] = preprocessing_labels[label]

        # sort model components
        sorted_model_labels = dict(sorted(model_labels.items(), key=lambda item: item[1], reverse=True))

        return sorted_preprocessing_labels, sorted_model_labels

    def _sort(self, preprocessing_set, label_order):
        n = len(preprocessing_set)

        # Traverse through all array elements
        for i in range(n - 1):
            # range(n) also work but outer loop will repeat one time more than needed.

            # Last i elements are already in place
            for j in range(0, n - i - 1):
                # traverse the array from 0 to n-i-1
                # Swap if the element found is greater
                # than the next element
                combination = preprocessing_set[j + 1] + "#" + preprocessing_set[j]

                if combination in label_order:
                    # logger.debug('combination', combination)
                    preprocessing_set[j], preprocessing_set[j + 1] = preprocessing_set[j + 1], preprocessing_set[j]
        return preprocessing_set

    def _get_adaptation_metric_label(self) -> Optional[str]:
        if self.task.adaptation_metric:
            metric = self.task.adaptation_metric.lower()
            if metric == "f1":
                return macros.Metric.F1.value
            elif metric == "r2":
                return macros.Metric.R2.value
            elif metric == "rmsle":
                return macros.Metric.RMSLE.value
            elif metric == "rmse":
                return macros.Metric.RMSE.value
            elif metric == "auc":
                return macros.Metric.AUC.value
            elif metric == "accuracy":
                return macros.Metric.Accuracy.value
            elif metric == "mae":
                return macros.Metric.MAE.value
            elif metric == "gini":
                return macros.Metric.Gini.value
            elif metric == "logloss":
                return macros.Metric.LogLoss.value
            elif metric == "roc_auc":
                return macros.Metric.ROC_AUC.value
            elif metric == "mcc":
                return macros.Metric.MCC.value
            elif metric == "qwk":
                return macros.Metric.QWK.value
            elif metric.startswith("map_"):
                k = metric.split("_")[1]
                try:
                    if int(k) < 1:
                        raise ValueError("Please set 'MAP_K' to a number greater than or equal to 1.")
                except Exception:
                    raise ValueError("Please set number for K in MAP_K.")
                return macros.Metric.MAP_K.value + k
            elif metric == "mape":
                return macros.Metric.MAPE.value
            else:
                raise ValueError(f"Invalid metric: {metric}")
        else:
            return None

    def _setup_pipeline_basics(self):
        pipeline = self.pipeline.pipeline
        pipeline.adaptation_metric = self.adaptation_metric
        pipeline.is_multi_class_multi_targets = (
            len(self.task.target_columns) > 1 and self.dataset_summary.has_multi_class_targets
        )

        # populate all columns and their datatypes
        all_columns_datatypes = dict()
        for column_name, column in self.dataset_summary.columns.items():
            all_columns_datatypes[column_name] = column.dtype
        pipeline.all_columns_datatypes = all_columns_datatypes
        pipeline.train_column_names = list(self.dataset_summary.columns.keys())
        pipeline.test_column_names = list(self.dataset_summary.columns.keys())

    def _populate_preprocessing_components_in_pipeline(self, preprocessing_components):
        if self.pipeline is None:
            return

        columns = self.dataset_summary.columns

        for component in preprocessing_components:
            # handle special case for SMOTE, don't apply if target columns are > 1. SMOTE fails in such cases
            if "PREPROCESS:Balancing:SMOTE:imblearn" == component.label_name and len(self.task.target_columns) > 1:
                continue

            rel_cols = component.get_relevant_columns(
                self.dataset_summary, self.task.target_columns, self.dataset_summary.cols_str_other
            )

            if "PREPROCESS:MissingValues:fillna" in component.label_name:
                bool_cols = [column_name for column_name, column in columns.items() if "bool" in str(column.dtype)]
                datetime_cols = [
                    column_name for column_name, column in columns.items() if "datetime" in str(column.dtype)
                ]
                # remove boolean and datetime columns, since Simple Imputer cannot handle boolean and datetime dtype
                rel_cols = sorted(list(set(rel_cols) - set(bool_cols) - set(datetime_cols)))
                rel_cols = sorted(
                    list(set(rel_cols) - set(self.dataset_summary.cols_str_other) - set(self.task.target_columns))
                )

            # handle special case for log transformation if target feature is in relevant columns
            if (
                is_allowed_to_apply_to_target(component.label_name)
                and len(list(set(rel_cols) & set(self.task.target_columns))) > 0
            ):
                if self.task.task_type == macros.TASK_REGRESSION:
                    # add code for inversing log transformation, if scaling on target feature
                    self.pipeline.pipeline.inverse_target = True
                else:
                    # in a classification task, an error is caused when the target is made continuous with log transformation
                    # hence, remove target from relevant columns in this scenario
                    rel_cols = sorted(list(set(rel_cols) - set(self.task.target_columns)))

            # construct component snippet
            if component.label_name in self.PREPROCESSING_AFTER_TARGET_SEPARATION:
                # remove target feature
                rel_cols = sorted(list(set(rel_cols) - set(self.task.target_columns)))

            # store relevant columns in the component after doing some additional processing, if needed
            component.relevant_columns = rel_cols

            # add component to pipeline
            if component.label_name in self.PREPROCESSING_AFTER_TARGET_SEPARATION:
                self.pipeline.add_processing_components(
                    component,
                    type="preprocessing_after_target_separation",
                    training_dataframe="feature_train",
                    test_dataframe="feature_test",
                )
            elif component.label_name in self.PREPROCESSING_AFTER_TRAIN_TEST_SPLIT:
                self.pipeline.add_processing_components(
                    component,
                    type="preprocessing_after_train_test_split",
                    training_dataframe="feature_train",
                    test_dataframe="feature_test",
                )
            else:
                self.pipeline.add_processing_components(
                    component,
                    type="preprocessing_before_target_separation",
                    training_dataframe="train_dataset",
                    test_dataframe="test_dataset",
                )

            # The information is used, for example when Scaler is used and Tfidf is not used.
            # We'd like to change numpy.ndarray of Scaler's output into pandas.DataFrame.
            if "PREPROCESS:Text:TfidfVectorizer" in component.label_name:
                self.pipeline.pipeline.sparse_matrix = True

            id = 1
            for component in (
                list(self.pipeline.pipeline.pipeline_json["preprocessing_before_target_separation"].values())
                + list(self.pipeline.pipeline.pipeline_json["preprocessing_after_target_separation"].values())
                + list(self.pipeline.pipeline.pipeline_json["preprocessing_after_train_test_split"].values())
            ):
                component["id"] = id
                id += len(component["code"])

    def _populate_model_components_in_pipeline(self, models: list[ModelLabel]) -> list[PipelineTemplate]:
        pipeline_list: list[PipelineTemplate] = []
        for model_obj in models:
            if self.pipeline.pipeline.adaptation_metric is not None and (
                self.pipeline.pipeline.adaptation_metric in macros.metric_needing_predict_proba
                or self.pipeline.pipeline.adaptation_metric.startswith(macros.Metric.MAP_K.value)
                or self.config.predict_option == macros.PRED_PROBABILITY
            ):
                model_obj.predict_proba = True

            # first variant with default hyperparameters
            pipeline_variant = copy.deepcopy(self.pipeline)
            pipeline_variant.pipeline.model = model_obj
            pipeline_list.append(pipeline_variant)

        return pipeline_list

    def run_adaptation(self) -> list[SimplePipeline]:
        """Create candidate pipelines.

        Returns
        -------
        list[SimplePipeline]

        """
        # SapientML default run
        self._setup_pipeline_basics()
        preprocessing, model = self._get_labels_from_skeleton_predictor()
        self._populate_preprocessing_components_in_pipeline(preprocessing)
        pipeline_list = self._populate_model_components_in_pipeline(model)
        for pipeline in pipeline_list:
            pipeline.generate()
            pipeline.pipeline.labels = self.labels

        return [SimplePipeline(**pipeline.pipeline.model_dump()) for pipeline in pipeline_list]
